"""
This module contains the YAMLSource class, a class for a YAML configuration
source.

This module is part of the `configorm` package for handling configuration data.
"""

from pathlib import Path
import yaml
from py_configorm.exception import ConfigORMError
from py_configorm.sources.base import SourceBase


class YAMLSource(SourceBase):
    """
    Class for a YAML configuration source.

    This class is a subclass of `SourceBase` and represents a YAML configuration
    source. It provides methods to load and save configuration data from/to a
    YAML file.

    Attributes:
        _file_path (Path): The path to the YAML configuration file.

    Methods:
        __init__(self, file_path: Path, readonly: bool = True):
            Initializes a new instance of `YAMLSource`.

        load(self) -> dict:
            Load configuration data from this source.

            Returns:
                dict: The loaded configuration data.

        save(self, data: dict):
            Save configuration data to this source.

            Args:
                data (dict): The configuration data to save.

        reload(self):
            Reload configuration data from this source.

            This method is called when the application is reloaded and the
            configuration data must be reloaded from the source.
    """

    def __init__(self, file_path: Path, readonly: bool = True):
        super().__init__(readonly)
        self._file_path = file_path

    def load(self) -> dict:
        """
        Load configuration data from this source.

        Returns:
            dict: The loaded configuration data.
        """
        try:
            with open(self._file_path, "r") as f:
                return yaml.safe_load(f)
        except FileNotFoundError:
            raise
        except yaml.YAMLError as e:
            raise ConfigORMError("Error loading YAML file: {}", format(e))

    def save(self, data: dict):
        """
        Save configuration data to this source.

        Args:
            data (dict): The configuration data to save.
        """
        try:
            if self.readonly:
                raise PermissionError("This source is read-only.")
            
            with open(self._file_path, "w") as f:
                yaml.dump(data, f)
        except (FileNotFoundError, PermissionError):
            raise



"""
This module contains the JSONSource class, a class for a JSON configuration
source.

Attributes:
    JSONSource (JSONSource): The JSONSource class.

"""

import json
from pathlib import Path

from py_configorm.exception import ConfigORMError
from py_configorm.sources.base import SourceBase


class JSONSource(SourceBase):
    """A class for a JSON configuration source.

    Attributes:
        _file_path (Path): The path to the JSON configuration file.

    Methods:
        __init__(self, file_path: Path, readonly: bool = True):
            Initializes a new instance of `JSONSource`.

        load(self) -> dict:
            Load configuration data from this source.

            Returns:
                dict: The loaded configuration data.

        save(self, data: dict):
            Save configuration data to this source.

            Args:
                data (dict): The configuration data to save.

        reload(self):
            Reload configuration data from this source.

            This method is called when the application is reloaded and the
            configuration data must be reloaded from the source.
    """

    def __init__(self, file_path: Path, readonly: bool = True):
        super().__init__(readonly)
        self._file_path = file_path

    def load(self) -> dict:
        """
        Load configuration data from this source.

        This method loads the configuration data from the JSON file specified
        during the initialization of this class.

        Returns:
            dict: The loaded configuration data.

        Raises:
            FileNotFoundError: If the specified JSON file does not exist.

            json.JSONDecodeError: If there is an error decoding the JSON data.

        """
        try:
            with open(self._file_path, "r") as f:
                return json.load(f)
        except FileNotFoundError:
            raise
        except json.JSONDecodeError as e:
            raise ConfigORMError("Error loading JSON file: {}", format(e))

    def save(self, data: dict):
        """
        Save configuration data to this source.

        This method saves the configuration data to the JSON file specified
        during the initialization of this class.

        Args:
            data (dict): The configuration data to save.

        Raises:
            PermissionError: If the source is read-only.
        """
        try:
            if self.readonly:
                raise PermissionError("This source is read-only.")

            with open(self._file_path, "w") as f:
                json.dump(data, f)
        except FileNotFoundError:
            raise
        except json.JSONDecodeError as e:
            raise ConfigORMError("Error saving JSON file: {}", format(e))


"""
ENVSource: A class for a environment variable configuration source.

This module contains the ENVSource class, a class for a environment variable
configuration source. It provides methods to load and save configuration data
from/to environment variables.

Attributes:
    ENVSource (ENVSource): The ENVSource class.

"""

import os

from py_configorm.sources.base import SourceBase


class ENVSource(SourceBase):
    """
    Class for a environment variable configuration source.

    This class is a subclass of `SourceBase` and represents a environment variable
    configuration source. It provides methods to load and save configuration data
    from/to environment variables.

    Environment variables are defined as per `[PREFIX]_[KEY][NESTING_SLUG][SUBKEY] = [VALUE]`
    format. For example, if `PREFIX` is set to `CFGORM` and `NESTING_SLUG` is
    set to `__`, then environment variables will be defined as 
    `CFGORM_[KEY]__[SUBKEY] = VALUE`. For providing bidirectional mutation operations
    for ORM, these variables need to be converted to a dictionary following the 
    semantics of the user specified by subclassing [py_configorm.core.ConfigSchema][].

    Attributes:
        prefix (str): The prefix for environment variables.
        nesting_slug (str): The string used to determine nesting in environment variables.

    Methods:
        __init__(self, prefix: str = "CFGORM", readonly: bool = True, nesting_slug: str = "__"):
            Initializes a new instance of `ENVSource`.

        load(self) -> dict:
            Load configuration data from this source.

            Returns:
                dict: The loaded configuration data.

        save(self, data: dict):
            Save configuration data to this source.

            Args:
                data (dict): The configuration data to save.

        reload(self):
            Reload configuration data from this source.

            This method is called when the application is reloaded and the
            configuration data must be reloaded from the source.
    """

    def __init__(
        self, prefix: str = "CFGORM_", readonly: bool = True, nesting_slug: str = "__"
    ):
        super().__init__(readonly)
        self._prefix = prefix
        self._nesting_slug = nesting_slug

    def load(self) -> dict:
        """Load configuration data from this source.

        Returns:
            dict: The loaded configuration data.
        """
        vars_ = {
            k[len(self._prefix) :]: v
            for k, v in os.environ.items()
            if k.startswith(self._prefix)
        }

        data = {}
        for key, value in vars_.items():
            l1_keys = key.split(self._nesting_slug, maxsplit=1)
            if l1_keys[0] in data:
                if l1_keys[1] in data[l1_keys[0]]:
                    data[l1_keys[0]][l1_keys[1]] = value
                else:
                    data[l1_keys[0]][l1_keys[1]] = {}
                    data[l1_keys[0]][l1_keys[1]] = value
            else:
                data[l1_keys[0]] = {}
                data[l1_keys[0]][l1_keys[1]] = value

        return data

    def save(self, data: dict):
        """Save configuration data to this source.

        Args:
            data (dict): The configuration data to save.
        """
        if self.readonly:
            raise PermissionError("This source is read-only.")
        
        for k0, v0 in data.items():
            if isinstance(v0, dict):
                for k1, v1 in v0.items():
                    if isinstance(v1, dict):
                        raise ValueError("Only one level of nesting is supported.")
                    os.environ[self._prefix + k0 + self._nesting_slug + k1] = v1
            else:
                os.environ[self._prefix + k0] = v0



"""
ConfigORM - A simple configuration library.

This module contains the ConfigORM class, a class for handling configuration data.

Classes:
    ConfigSchema (ConfigSchema): The ConfigSchema class.
    ConfigORM (ConfigORM): The ConfigORM class.
"""

from typing import List, Type
from pydantic import BaseModel
import pydantic

from py_configorm.exception import ConfigORMError


class ConfigSchema(BaseModel):
    """
    Base class for all user defined configuration schemas.
    """

    pass


class ConfigORM:
    def __init__(self, schema: Type[ConfigSchema], sources: List):
        self._schema = schema
        self._sources = sources
        self._config = self.load_config()

    def load_config(self) -> ConfigSchema:
        """
        Load configuration data from all the sources.

        This method loads the configuration data from all the sources specified
        during the initialization of this class. The configuration data is
        merged together and returned as a single `ConfigSchema` object.

        Returns:
            ConfigSchema: The loaded configuration data.
        """
        config_data = {}
        try:

            def merge_config(data, new_data):
                for key, value in new_data.items():
                    if isinstance(value, dict) and key in data:
                        merge_config(data[key], value)
                    else:
                        data[key] = value

            if len(self._sources) == 0:
                raise ConfigORMError("No configuration sources specified")

            for source in self._sources:
                merge_config(config_data, source.load())

            return self._schema(**config_data)
        except Exception as e:
            raise ConfigORMError("Error loading configuration data: {}", format(e))
        except pydantic.ValidationError as e:
            raise ConfigORMError("Invalid configuration data: {}", format(e))
        except TypeError as e:
            raise ConfigORMError("Invalid configuration data type: {}", format(e))

    def save_config(self):
        """
        Save configuration data to all the sources.

        This method saves the configuration data to all the sources specified
        during the initialization of this class.

        Raises:
            PermissionError: If one of the sources is read-only.
        """

        if len(self._sources) == 0:
            raise ConfigORMError("No configuration sources specified")

        for source in self._sources:
            if not source.readonly:
                source.save(self._config.model_dump())

    def reload_config(self):
        """
        Reload configuration data from all the sources.

        This method reloads the configuration data from all the sources specified
        during the initialization of this class. The configuration data is
        merged together and returned as a single `ConfigSchema` object.
        """
        self._config = self.load_config()



"""
ConfigORM - A simple configuration library.

This module contains the DOTENVSource class, a class for a dotenv configuration
source.

Attributes:
    DOTENVSource (DOTENVSource): The DOTENVSource class.

"""

from pathlib import Path

import dotenv

from py_configorm.sources.base import SourceBase


class DOTENVSource(SourceBase):
    """Class for a dotenv configuration source.

    This class is a subclass of `SourceBase` and represents a dotenv
    configuration source. It provides methods to load and save configuration
    data from/to a dotenv file.

    Attributes:
        _file_path (Path): The path to the dotenv configuration file.

    Methods:
        __init__(self, file_path: Path, readonly: bool = True):
            Initializes a new instance of `DOTENVSource`.

        load(self) -> dict:
            Load configuration data from this source.

            Returns:
                dict: The loaded configuration data.

        save(self, data: dict):
            Save configuration data to this source.

            Args:
                data (dict): The configuration data to save.

        reload(self):
            Reload configuration data from this source.

            This method is called when the application is reloaded and the
            configuration data must be reloaded from the source.

    """

    def __init__(
        self,
        file_path: Path,
        prefix: str = "CFGORM_",
        readonly: bool = True,
        nesting_slug: str = "__",
    ):
        super().__init__(readonly)
        self._file_path = file_path
        self._prefix = prefix
        self._nesting_slug = nesting_slug

    def load(self) -> dict:
        """Load configuration data from this source.

        This method loads the configuration data from the dotenv file specified
        during the initialization of this class.

        Returns:
            dict: The loaded configuration data.

        Raises:
            FileNotFoundError: If the specified dotenv file does not exist.
        """
        try:
            data = {}
            with open(self._file_path, "r") as f:
                vars_ = dotenv.dotenv_values(dotenv_path=f.name)
                for key, value in vars_.items():
                    n_key = key.split(self._prefix, maxsplit=1)[1]
                    l1_keys = n_key.split(self._nesting_slug, maxsplit=1)
                    if l1_keys[0] in data:
                        if l1_keys[1] in data[l1_keys[0]]:
                            data[l1_keys[0]][l1_keys[1]] = value
                        else:
                            data[l1_keys[0]][l1_keys[1]] = {}
                            data[l1_keys[0]][l1_keys[1]] = value
                    else:
                        data[l1_keys[0]] = {}
                        data[l1_keys[0]][l1_keys[1]] = value
                return data
        except FileNotFoundError:
            raise

    def save(self, data: dict):
        """Save configuration data to this source.

        This method saves the configuration data to the dotenv file specified
        during the initialization of this class.

        Args:
            data (dict): The configuration data to save.

        Raises:
            PermissionError: If the source is read-only.
        """
        if self.readonly:
            raise PermissionError("This source is read-only.")

        raise NotImplementedError("This source does not support saving.")



"""
ConfigORM - A simple configuration library.

This is a python library for handling configuration data.

"""

from abc import ABC, abstractmethod


class SourceBase(ABC):
    """
    Base class for all configuration sources.

    This class is an abstract base class for all configuration sources.
    It defines the basic interface for loading and saving configuration
    data.

    Attributes:
        readonly (bool): Whether the source is read-only.
    """

    def __init__(self, readonly: bool = True):
        self._readonly = readonly

    @abstractmethod
    def load(self) -> dict:
        """
        Load configuration data from this source.

        Returns:
            dict: The loaded configuration data.
        """
        pass

    @abstractmethod
    def save(self, data: dict):
        """
        Save configuration data to this source.

        Args:
            data (dict): The configuration data to save.
        """
        pass

    # @abstractmethod
    # def reload(self):
    #     """
    #     Reload configuration data from this source.

    #     This method is called when the application is reloaded and the
    #     configuration data must be reloaded from the source.

    #     """
    #     pass

    @property
    def readonly(self) -> bool:
        return self._readonly
